.labelValue {
  cursor: pointer;
}

.labelValue:hover {
  background-color: light-dark(
    var(--mantine-color-gray-2),
    var(--mantine-color-gray-8)
  );
  border-radius: var(--mantine-radius-sm);
}

.promqlPill {
  background-color: light-dark(
    var(--mantine-color-gray-1),
    var(--mantine-color-dark-5)
  );
}
