.codebox {
  background-color: light-dark(
    var(--mantine-color-gray-1),
    var(--mantine-color-gray-9)
  );
}

.queryButton {
  opacity: 0;
  transition: opacity 0.1s ease-in-out;
}

.codebox:hover .queryButton {
  opacity: 1;
}
