// Copyright 2024 The Prometheus Authors
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

//go:build linux && !forcedirectio

package fileutil

import (
	"bufio"
	"os"
)

func NewBufioWriterWithSeek(f *os.File, size int) (BufWriter, error) {
	return &writer{bufio.NewWriterSize(f, size)}, nil
}

func NewDirectIOWriter(f *os.File, size int) (BufWriter, error) {
	return newDirectIOWriter(f, size)
}
