.tableWrapper {
  border: 1px solid
    light-dark(var(--mantine-color-gray-3), var(--mantine-color-dark-5));
  border-radius: var(--mantine-radius-default);
}

.numberCell {
  text-align: right;
  font-variant-numeric: tabular-nums;
}
