// Copyright 2025 The Prometheus Authors
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package main

import (
	"flag"
	"fmt"
	"os"

	"github.com/prometheus/prometheus/promql/promqltest"
)

func main() {
	mode := flag.String("mode", "strict", "Migration mode: strict, basic, or tolerant")
	dir := flag.String("dir", "", "Directory to migrate")
	flag.Parse()

	if err := promqltest.MigrateTestData(*mode, *dir); err != nil {
		fmt.Printf("Error migrating test files: %v\n", err)
		os.Exit(1)
	}
}
